# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from glom import glom
import six


def int_or_none(data):
    try:
        return int(data)
    except (TypeError, ValueError):
        return None


def convert_to_crash_data(raw_crash, processed_crash):
    """
    Takes a raw crash and a processed crash (these are Socorro-centric
    data structures) and converts them to a crash data structure used
    by signature generation.

    :arg raw_crash: raw crash data from Socorro
    :arg processed_crash: processed crash data from Socorro

    :returns: crash data structure that conforms to the schema

    """
    # We want to generate fresh signatures, so we remove the "normalized" field
    # from stack frames from the processed crash because this is essentially
    # cached data from previous processing
    for thread in glom(processed_crash, 'json_dump.threads', default=[]):
        for frame in thread.get('frames', []):
            if 'normalized' in frame:
                del frame['normalized']

    crash_data = {
        # JavaStackTrace or None
        'java_stack_trace': glom(raw_crash, 'JavaStackTrace', default=None),

        # int or None
        'crashing_thread': glom(
            processed_crash, 'json_dump.crash_info.crashing_thread', default=None
        ),

        # list of CStackTrace or None
        'threads': glom(processed_crash, 'json_dump.threads', default=None),

        # int or None
        'hang_type': glom(processed_crash, 'hang_type', default=None),

        # text or None
        'os': glom(processed_crash, 'json_dump.system_info.os', default=None),

        # int or None
        'oom_allocation_size': int_or_none(glom(raw_crash, 'OOMAllocationSize', default=None)),

        # text or None
        'abort_message': glom(raw_crash, 'AbortMessage', default=None),

        # text or None
        'mdsw_status_string': glom(processed_crash, 'mdsw_status_string', default=None),

        # text json with "phase", "conditions" (complicated--see code) or None
        'async_shutdown_timeout': glom(raw_crash, 'AsyncShutdownTimeout', default=None),

        # text or None
        'jit_category': glom(processed_crash, 'classifications.jit.category', default=None),

        # text or None
        'ipc_channel_error': glom(raw_crash, 'ipc_channel_error', default=None),

        # text or None
        'ipc_message_name': glom(raw_crash, 'IPCMessageName', default=None),

        # text
        'moz_crash_reason': glom(raw_crash, 'MozCrashReason', default=None),

        # text; comma-delimited e.g. "browser,flash1,flash2"
        'additional_minidumps': glom(raw_crash, 'additional_minidumps', default=''),

        # pull out the original signature if there was one
        'original_signature': glom(processed_crash, 'signature', default='')
    }
    return crash_data


#: List of allowed characters: ascii, printable, and non-whitespace except space
ALLOWED_CHARS = [chr(c) for c in range(32, 127)]


def drop_bad_characters(text):
    """Takes a text and drops all non-printable and non-ascii characters and
    also any whitespace characters that aren't space.

    :arg str/unicode text: the text to fix

    :returns: text with all bad characters dropped

    """
    if six.PY2:
        if isinstance(text, str):
            # Convert to unicode to handle encoded sequences
            text = text.decode('unicode_escape')
        # Convert to a Python 2 str and drop any non-ascii characters
        text = text.encode('ascii', 'ignore')

    # Strip all non-ascii and non-printable characters
    text = ''.join([c for c in text if c in ALLOWED_CHARS])
    return text



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import importlib

import mock

from . import WHATEVER

# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
base_module = '.'.join(__name__.split('.')[:-2])
generator = importlib.import_module(base_module + '.generator')


class TestSignatureGenerator:
    def test_empty_dicts(self):
        generator_obj = generator.SignatureGenerator()
        ret = generator_obj.generate({})

        # NOTE(willkg): This is what the current pipeline yields. If any of those parts change, this
        # might change, too. The point of this test is that we can pass in empty dicts and the
        # SignatureGenerator and the rules in the default pipeline don't fall over.
        expected = {
            'notes': [
                'CSignatureTool: No signature could be created because we do not know '
                'which thread crashed'
            ],
            'signature': 'EMPTY: no crashing thread identified'
        }

        assert ret == expected

    def test_failing_rule(self):
        class BadRule(object):
            pass

        generator_obj = generator.SignatureGenerator(pipeline=[BadRule()])
        ret = generator_obj.generate({})

        expected = {
            'notes': [
                'Rule BadRule failed: \'BadRule\' object has no attribute \'predicate\''
            ],
            'signature': ''
        }

        assert ret == expected

    def test_error_handler(self):
        exc_value = Exception('Cough')

        class BadRule(object):
            def predicate(self, crash_data, result):
                raise exc_value

        error_handler = mock.MagicMock()

        generator_obj = generator.SignatureGenerator(
            pipeline=[BadRule()], error_handler=error_handler
        )
        generator_obj.generate({'uuid': 'ou812'})

        # Make sure error_handler was called with right extra
        assert (
            error_handler.call_args_list == [
                mock.call(
                    {'uuid': 'ou812'},
                    exc_info=(Exception, exc_value, WHATEVER),
                    extra={'rule': 'BadRule'}
                )
            ]
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from itertools import islice
import re

from glom import glom
import ujson

from . import siglists_utils
from .utils import drop_bad_characters


SIGNATURE_MAX_LENGTH = 255
MAXIMUM_FRAMES_TO_CONSIDER = 40


class Rule(object):
    """Base class for Signature generation rules"""
    def predicate(self, crash_data, result):
        """Whether or not to run this rule

        :arg dict crash_data: the data to use to generate the signature
        :arg dict result: the current signature generation result

        :returns: True or False

        """
        return True

    def action(self, crash_data, result):
        """Runs the rule against the data

        .. Note::

           This modifies ``result`` in place.

        :arg dict crash_data: the data to use to generate the signature
        :arg dict result: the current signature generation result

        :returns: True

        """
        return True


class SignatureTool(object):
    """Stack walking signature generator base class

    This defines the interface for classes that take a stack and generate a
    signature from it.

    Subclasses should implement the ``_do_generate`` method.

    """

    def __init__(self, quit_check_callback=None):
        self.quit_check_callback = quit_check_callback

    def generate(self, source_list, hang_type=0, crashed_thread=None, delimiter=' | '):
        signature, signature_notes = self._do_generate(
            source_list,
            hang_type,
            crashed_thread,
            delimiter
        )

        return signature, signature_notes

    def _do_generate(self, source_list, hang_type, crashed_thread, delimiter):
        raise NotImplementedError


class CSignatureTool(SignatureTool):
    """This is the class for signature generation tools that work on breakpad C/C++
    stacks. It provides a method to normalize signatures and then defines its
    own '_do_generate' method.

    """

    hang_prefixes = {
        -1: "hang",
        1: "chromehang"
    }

    def __init__(self, quit_check_callback=None):
        super(CSignatureTool, self).__init__(quit_check_callback)

        self.irrelevant_signature_re = re.compile(
            '|'.join(siglists_utils.IRRELEVANT_SIGNATURE_RE)
        )
        self.prefix_signature_re = re.compile(
            '|'.join(siglists_utils.PREFIX_SIGNATURE_RE)
        )
        self.signatures_with_line_numbers_re = re.compile(
            '|'.join(siglists_utils.SIGNATURES_WITH_LINE_NUMBERS_RE)
        )
        self.trim_dll_signature_re = re.compile(
            '|'.join(siglists_utils.TRIM_DLL_SIGNATURE_RE)
        )
        self.signature_sentinels = siglists_utils.SIGNATURE_SENTINELS

        self.collapse_arguments = True

        self.fixup_space = re.compile(r' (?=[\*&,])')
        self.fixup_comma = re.compile(r',(?! )')
        self.fixup_hash = re.compile(r'::h[0-9a-fA-F]+$')

    @staticmethod
    def _is_exception(exception_list, remaining_original_line, line_up_to_current_position):
        for an_exception in exception_list:
            if remaining_original_line.startswith(an_exception):
                return True
            if line_up_to_current_position.endswith(an_exception):
                return True
        return False

    def _collapse(
        self,
        function_signature_str,
        open_string,
        replacement_open_string,
        close_string,
        replacement_close_string,
        exception_substring_list=[],
    ):
        """this method takes a string representing a C/C++ function signature
        and replaces anything between to possibly nested delimiters

        :arg list exception_substring_list: list of exceptions that shouldn't collapse

        """
        target_counter = 0
        collapsed_list = []
        exception_mode = False

        def append_if_not_in_collapse_mode(a_character):
            if not target_counter:
                collapsed_list.append(a_character)

        for index, a_character in enumerate(function_signature_str):
            if a_character == open_string:
                if self._is_exception(
                    exception_substring_list,
                    function_signature_str[index + 1:],
                    function_signature_str[:index]
                ):
                    exception_mode = True
                    append_if_not_in_collapse_mode(a_character)
                    continue
                append_if_not_in_collapse_mode(replacement_open_string)
                target_counter += 1
            elif a_character == close_string:
                if exception_mode:
                    append_if_not_in_collapse_mode(a_character)
                    exception_mode = False
                else:
                    target_counter -= 1
                    append_if_not_in_collapse_mode(replacement_close_string)
            else:
                append_if_not_in_collapse_mode(a_character)

        edited_function = ''.join(collapsed_list)
        return edited_function

    def normalize_signature(
        self,
        module=None,
        function=None,
        file=None,
        line=None,
        module_offset=None,
        offset=None,
        normalized=None,
        **kwargs  # eat any extra kwargs passed in
    ):
        """Normalizes a single frame into a signature part

        Returns a structured conglomeration of the input parameters to serve as
        a signature. The parameter names of this function reflect the exact
        names of the fields from the jsonMDSW frame output. This allows this
        function to be invoked by passing a frame as ``**a_frame``.

        Sometimes, a frame may already have a normalized version cached. If
        that exists, return it instead.

        """
        # If there's a cached normalized value, use that so we don't spend time
        # figuring it out again
        if normalized is not None:
            return normalized

        # If there's a function (and optionally line), use that
        if function:
            function = self._collapse(
                function,
                '<',
                '<',
                '>',
                'T>',
                ('name omitted', 'IPC::ParamTraits')
            )
            if self.collapse_arguments:
                function = self._collapse(
                    function,
                    '(',
                    '',
                    ')',
                    '',
                    ('anonymous namespace', 'operator')
                )
            if 'clone .cold' in function:
                # Remove PGO cold block labels like "[clone .cold.222]". bug #1397926
                function = self._collapse(
                    function,
                    '[',
                    '',
                    ']',
                    ''
                )
            if self.signatures_with_line_numbers_re.match(function):
                function = "%s:%s" % (function, line)
            # Remove spaces before all stars, ampersands, and commas
            function = self.fixup_space.sub('', function)
            # Ensure a space after commas
            function = self.fixup_comma.sub(', ', function)
            # Remove rust-generated uniqueness hashes
            function = self.fixup_hash.sub('', function)
            return function

        # If there's a file and line number, use that
        if file and line:
            filename = file.rstrip('/\\')
            if '\\' in filename:
                file = filename.rsplit('\\')[-1]
            else:
                file = filename.rsplit('/')[-1]
            return '%s#%s' % (file, line)

        # If there's an offset and no module/module_offset, use that
        if not module and not module_offset and offset:
            return "@%s" % offset

        # Return module/module_offset
        return '%s@%s' % (module or '', module_offset)

    def _do_generate(self, source_list, hang_type, crashed_thread, delimiter=' | '):
        """
        each element of signatureList names a frame in the crash stack; and is:
          - a prefix of a relevant frame: Append this element to the signature
          - a relevant frame: Append this element and stop looking
          - irrelevant: Append this element only after seeing a prefix frame
        The signature is a ' | ' separated string of frame names.
        """
        signature_notes = []

        # shorten source_list to the first signatureSentinel
        sentinel_locations = []
        for a_sentinel in self.signature_sentinels:
            if type(a_sentinel) == tuple:
                a_sentinel, condition_fn = a_sentinel
                if not condition_fn(source_list):
                    continue
            try:
                sentinel_locations.append(source_list.index(a_sentinel))
            except ValueError:
                pass
        if sentinel_locations:
            source_list = source_list[min(sentinel_locations):]

        # Get all the relevant frame signatures.
        new_signature_list = []
        for a_signature in source_list:
            # We want to match against the function signature or any of the parts of the
            # signature in the case where there are specifiers and return types.
            a_signature_parts = a_signature.split() + [a_signature]

            # If one of the parts of the function signature matches the
            # irrelevant signatures regex, skip to the next frame.
            if any(self.irrelevant_signature_re.match(part) for part in a_signature_parts):
                continue

            # If the signature matches the trim dll signatures regex, rewrite it to remove all but
            # the module name.
            if self.trim_dll_signature_re.match(a_signature):
                a_signature = a_signature.split('@')[0]

                # If this trimmed DLL signature is the same as the previous frame's, we do not want
                # to add it.
                if new_signature_list and a_signature == new_signature_list[-1]:
                    continue

            new_signature_list.append(a_signature)

            # If none of the parts of the function signature signature matches
            # the prefix signatures regex, then it is the last one we add to
            # the list.
            if not any(self.prefix_signature_re.match(part) for part in a_signature_parts):
                break

        # Add a special marker for hang crash reports.
        if hang_type:
            new_signature_list.insert(0, self.hang_prefixes[hang_type])

        signature = delimiter.join(new_signature_list)

        # Handle empty signatures to explain why we failed generating them.
        if signature == '' or signature is None:
            if crashed_thread is None:
                signature_notes.append(
                    "CSignatureTool: No signature could be created because we do not know which "
                    "thread crashed"
                )
                signature = "EMPTY: no crashing thread identified"
            else:
                signature_notes.append(
                    "CSignatureTool: No proper signature could be created because no good data "
                    "for the crashing thread (%s) was found" % crashed_thread
                )
                try:
                    signature = source_list[0]
                except IndexError:
                    signature = "EMPTY: no frame data available"

        return signature, signature_notes


class JavaSignatureTool(SignatureTool):
    """This is the signature generation class for Java signatures."""

    # The max length of a java exception description--if it's longer than this,
    # drop it
    DESCRIPTION_MAX_LENGTH = 255

    java_line_number_killer = re.compile(r'\.java\:\d+\)$')
    java_hex_addr_killer = re.compile(r'@[0-9a-f]{8}')

    @staticmethod
    def join_ignore_empty(delimiter, list_of_strings):
        return delimiter.join(x for x in list_of_strings if x)

    def _do_generate(self, source, hang_type_unused=0, crashed_thread_unused=None, delimiter=': '):
        signature_notes = []
        try:
            source_list = [x.strip() for x in source.splitlines()]
        except AttributeError:
            signature_notes.append('JavaSignatureTool: stack trace not in expected format')
            return (
                "EMPTY: Java stack trace not in expected format",
                signature_notes
            )

        try:
            java_exception_class, description = source_list[0].split(':', 1)
            java_exception_class = java_exception_class.strip()
            # relace all hex addresses in the description by the string <addr>
            description = self.java_hex_addr_killer.sub(
                r'@<addr>',
                description
            ).strip()
        except ValueError:
            java_exception_class = source_list[0]
            description = ''
            signature_notes.append(
                'JavaSignatureTool: stack trace line 1 is not in the expected format'
            )

        try:
            java_method = re.sub(
                self.java_line_number_killer,
                '.java)',
                source_list[1]
            )
            if not java_method:
                signature_notes.append('JavaSignatureTool: stack trace line 2 is empty')
        except IndexError:
            signature_notes.append('JavaSignatureTool: stack trace line 2 is missing')
            java_method = ''

        # An error in an earlier version of this code resulted in the colon
        # being left out of the division between the description and the
        # java_method if the description didn't end with "<addr>". This code
        # perpetuates that error while correcting the "<addr>" placement when
        # it is not at the end of the description. See Bug 865142 for a
        # discussion of the issues.
        if description.endswith('<addr>'):
            # at which time the colon placement error is to be corrected
            # just use the following line as the replacement for this entire
            # if/else block
            signature = self.join_ignore_empty(
                delimiter,
                (java_exception_class, description, java_method)
            )
        else:
            description_java_method_phrase = self.join_ignore_empty(
                ' ',
                (description, java_method)
            )
            signature = self.join_ignore_empty(
                delimiter,
                (java_exception_class, description_java_method_phrase)
            )

        if len(signature) > self.DESCRIPTION_MAX_LENGTH:
            signature = delimiter.join(
                (java_exception_class, java_method)
            )
            signature_notes.append(
                'JavaSignatureTool: dropped Java exception description due to length'
            )

        return signature, signature_notes


class SignatureGenerationRule(Rule):

    def __init__(self):
        super(SignatureGenerationRule, self).__init__()
        self.java_signature_tool = JavaSignatureTool()
        self.c_signature_tool = CSignatureTool()

    def _create_frame_list(self, crashing_thread_mapping, make_modules_lower_case=False):
        frame_signatures_list = []
        for a_frame in islice(
            crashing_thread_mapping.get('frames', []),
            MAXIMUM_FRAMES_TO_CONSIDER
        ):
            if make_modules_lower_case and 'module' in a_frame:
                a_frame['module'] = a_frame['module'].lower()

            normalized_signature = self.c_signature_tool.normalize_signature(**a_frame)
            if 'normalized' not in a_frame:
                a_frame['normalized'] = normalized_signature
            frame_signatures_list.append(normalized_signature)
        return frame_signatures_list

    def _get_crashing_thread(self, crash_data):
        return crash_data.get('crashing_thread', 0)

    def action(self, crash_data, result):
        # If this is a Java crash, then generate a Java signature
        if crash_data.get('java_stack_trace'):
            signature, signature_notes = self.java_signature_tool.generate(
                crash_data['java_stack_trace'],
                delimiter=': '
            )
            result['signature'] = signature
            result['notes'].extend(signature_notes)
            return True

        try:
            # First, we need to figure out which thread to look at. If it's a
            # chrome hang (1), then use thread 0. Otherwise, use the crashing
            # thread specified in the crash data.
            if crash_data.get('hang_type', None) == 1:
                crashing_thread = 0
            else:
                crashing_thread = self._get_crashing_thread(crash_data)

            # If we have a thread to look at, pull the frames for that.
            # Otherwise we don't have frames to use.
            if crashing_thread is not None:
                signature_list = self._create_frame_list(
                    glom(crash_data, 'threads.%d' % crashing_thread, default={}),
                    crash_data.get('os') == 'Windows NT'
                )

            else:
                signature_list = []

        except (KeyError, IndexError) as exc:
            signature_notes.append('No crashing frames found because of %s' % exc)
            signature_list = []

        signature, signature_notes = self.c_signature_tool.generate(
            signature_list,
            crash_data.get('hang_type'),
            crash_data.get('crashing_thread')
        )

        if signature_list:
            result['proto_signature'] = ' | '.join(signature_list)
        result['signature'] = signature
        result['notes'].extend(signature_notes)

        return True


class OOMSignature(Rule):
    """To satisfy Bug 1007530, this rule will modify the signature to
    tag OOM (out of memory) crashes"""

    signature_fragments = (
        'NS_ABORT_OOM',
        'mozalloc_handle_oom',
        'CrashAtUnhandlableOOM',
        'AutoEnterOOMUnsafeRegion',
        'alloc::oom::oom',
    )

    def predicate(self, crash_data, result):
        if crash_data.get('oom_allocation_size'):
            return True

        signature = result['signature']
        if not signature:
            return False

        for a_signature_fragment in self.signature_fragments:
            if a_signature_fragment in signature:
                return True

        return False

    def action(self, crash_data, result):
        try:
            size = int(crash_data.get('oom_allocation_size'))
        except (TypeError, AttributeError, KeyError):
            result['signature'] = 'OOM | unknown | ' + result['signature']
            return True

        if size <= 262144:  # 256K
            result['signature'] = 'OOM | small'
        else:
            result['signature'] = 'OOM | large | ' + result['signature']
        return True


class AbortSignature(Rule):
    """Adds abort message data to the beginning of the signature

    See bug #803779.

    """

    def predicate(self, crash_data, result):
        return bool(crash_data.get('abort_message'))

    def action(self, crash_data, result):
        abort_message = crash_data['abort_message']

        if '###!!! ABORT: file ' in abort_message:
            # This is an abort message that contains no interesting
            # information. We just want to put the "Abort" marker in the
            # signature.
            result['signature'] = 'Abort | ' + result['signature']
            return True

        if '###!!! ABORT:' in abort_message:
            # Recent crash reports added some irrelevant information at the
            # beginning of the abort message. We want to remove that and keep
            # just the actual abort message.
            abort_message = abort_message.split('###!!! ABORT:', 1)[1]

        if ': file ' in abort_message:
            # Abort messages contain a file name and a line number. Since
            # those are very likely to change between builds, we want to
            # remove those parts from the signature.
            abort_message = abort_message.split(': file ', 1)[0]

        if 'unable to find a usable font' in abort_message:
            # "unable to find a usable font" messages include a parenthesized localized message. We
            # want to remove that. Bug #1385966
            open_paren = abort_message.find('(')
            if open_paren != -1:
                end_paren = abort_message.rfind(')')
                if end_paren != -1:
                    abort_message = abort_message[:open_paren] + abort_message[end_paren + 1:]

        abort_message = drop_bad_characters(abort_message).strip()

        if len(abort_message) > 80:
            abort_message = abort_message[:77] + '...'

        result['signature'] = 'Abort | %s | %s' % (abort_message, result['signature'])
        return True


class SigFixWhitespace(Rule):
    """Fix whitespace in signatures

    This does the following:

    * trims leading and trailing whitespace
    * converts all non-space whitespace characters to space
    * reduce consecutive spaces to a single space

    """

    WHITESPACE_RE = re.compile('\s')
    CONSECUTIVE_WHITESPACE_RE = re.compile('\s\s+')

    def predicate(self, crash_data, result):
        return isinstance(result.get('signature'), basestring)

    def action(self, crash_data, result):
        sig = result['signature']

        # Trim leading and trailing whitespace
        sig = sig.strip()

        # Convert all non-space whitespace characters into spaces
        sig = self.WHITESPACE_RE.sub(' ', sig)

        # Reduce consecutive spaces to a single space
        sig = self.CONSECUTIVE_WHITESPACE_RE.sub(' ', sig)

        result['signature'] = sig
        return True


class SigTruncate(Rule):
    """Truncates signatures down to SIGNATURE_MAX_LENGTH characters"""

    def predicate(self, crash_data, result):
        return len(result['signature']) > SIGNATURE_MAX_LENGTH

    def action(self, crash_data, result):
        max_length = SIGNATURE_MAX_LENGTH - 3
        result['signature'] = "%s..." % result['signature'][:max_length]
        result['notes'].append('SigTrunc: signature truncated due to length')
        return True


class StackwalkerErrorSignatureRule(Rule):
    """ensure that the signature contains the stackwalker error message"""

    def predicate(self, crash_data, result):
        return bool(
            result['signature'].startswith('EMPTY') and
            crash_data.get('mdsw_status_string')
        )

    def action(self, crash_data, result):
        result['signature'] = "%s; %s" % (
            result['signature'],
            crash_data['mdsw_status_string']
        )
        return True


class SignatureRunWatchDog(SignatureGenerationRule):
    """ensure that the signature contains the stackwalker error message"""

    def predicate(self, crash_data, result):
        return 'RunWatchdog' in result['signature']

    def _get_crashing_thread(self, crash_data):
        # Always use thread 0 in this case, because that's the thread that
        # was hanging when the software was artificially crashed.
        return 0

    def action(self, crash_data, result):
        # For shutdownhang crashes, we need to use thread 0 instead of the
        # crashing thread. The reason is because those crashes happen
        # artificially when thread 0 gets stuck. So whatever the crashing
        # thread is, we don't care about it and only want to know what was
        # happening in thread 0 when it got stuck.
        ret = super(SignatureRunWatchDog, self).action(crash_data, result)
        result['signature'] = (
            "shutdownhang | %s" % result['signature']
        )
        return ret


class SignatureShutdownTimeout(Rule):
    """replaces the signature if there is a shutdown timeout message in the
    crash"""

    def predicate(self, crash_data, result):
        return bool(crash_data.get('async_shutdown_timeout'))

    def action(self, crash_data, result):
        parts = ['AsyncShutdownTimeout']
        try:
            shutdown_data = ujson.loads(crash_data['async_shutdown_timeout'])
            parts.append(shutdown_data['phase'])
            conditions = [
                # NOTE(willkg): The AsyncShutdownTimeout notation condition can either be a string
                # that looks like a "name" or a dict with a "name" in it.
                #
                # This handles both variations.
                c['name'] if isinstance(c, dict) else c
                for c in shutdown_data['conditions']
            ]
            if conditions:
                conditions.sort()
                parts.append(','.join(conditions))
            else:
                parts.append("(none)")
        except (ValueError, KeyError) as exc:
            parts.append("UNKNOWN")
            result['notes'].append('Error parsing AsyncShutdownTimeout: {}'.format(exc))

        new_sig = ' | '.join(parts)
        result['notes'].append(
            'Signature replaced with a Shutdown Timeout signature, '
            'was: "{}"'.format(result['signature'])
        )
        result['signature'] = new_sig

        return True


class SignatureJitCategory(Rule):
    """replaces the signature if there is a JIT classification in the crash"""

    def predicate(self, crash_data, result):
        return bool(crash_data.get('jit_category'))

    def action(self, crash_data, result):
        result['notes'].append(
            'Signature replaced with a JIT Crash Category, '
            'was: "{}"'.format(result['signature'])
        )
        result['signature'] = "jit | {}".format(crash_data['jit_category'])
        return True


class SignatureIPCChannelError(Rule):
    """replaces the signature if there is a IPC channel error in the crash"""

    def predicate(self, crash_data, result):
        return bool(crash_data.get('ipc_channel_error'))

    def action(self, crash_data, result):
        if crash_data.get('additional_minidumps') == 'browser':
            new_sig = 'IPCError-browser | {}'
        else:
            new_sig = 'IPCError-content | {}'
        new_sig = new_sig.format(crash_data['ipc_channel_error'][:100])

        result['notes'].append(
            'Signature replaced with an IPC Channel Error, '
            'was: "{}"'.format(result['signature'])
        )
        result['signature'] = new_sig

        return True


class SignatureIPCMessageName(Rule):
    """augments the signature if there is a IPC message name in the crash"""

    def predicate(self, crash_data, result):
        return bool(crash_data.get('ipc_message_name'))

    def action(self, crash_data, result):
        result['signature'] = '{} | IPC_Message_Name={}'.format(
            result['signature'],
            crash_data['ipc_message_name']
        )
        return True


class SignatureParentIDNotEqualsChildID(Rule):
    """Stomp on the signature if MozCrashReason is parentBuildID != childBuildID

    In the case where the assertion fails, then the parent buildid and the child buildid are
    different. This causes a lot of strangeness particularly in symbolification, so the signatures
    end up as junk. Instead, we want to bucket all these together so we replace the signature.

    """

    def predicate(self, crash_data, result):
        value = 'MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)'
        return crash_data.get('moz_crash_reason') == value

    def action(self, crash_data, result):
        result['notes'].append(
            'Signature replaced with MozCrashAssert, was: "%s"' % result['signature']
        )

        # The MozCrashReason lists the assertion that failed, so we put "!=" in the signature
        result['signature'] = 'parentBuildID != childBuildID'

        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest

from ..utils import drop_bad_characters


@pytest.mark.parametrize('text, expected', [
    ('', ''),
    (u'', ''),

    ('123', '123'),
    (u'123', '123'),

    # Drop non-ascii characters
    ('1\xc6\x8a23', '123'),
    (u'1\u018a23', '123'),

    # Drop non-space whitespace characters
    ('\r\n\t1 23', '1 23'),
    (u'\r\n\t1 23', '1 23'),

    # Drop non-printable characters
    ('\0\b1 23', '1 23'),
    (u'\0\b1 23', '1 23'),
])
def test_drop_bad_characters(text, expected):
    assert drop_bad_characters(text) == expected



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import re

from pkg_resources import resource_stream
import six


# This is a hack because sentinels can be a tuple, with the second item being
# a function to verify if the sentinel applies. It's quite hard to express
# that in a .txt file, so this special value is here. This list should not
# grow much, and if it does, we should find a better solution for handling
# these special values.
_SPECIAL_EXTENDED_VALUES = {
    'signature_sentinels': [
        (
            'mozilla::ipc::RPCChannel::Call(IPC::Message*, IPC::Message*)',
            lambda x: (
                'CrashReporter::CreatePairedMinidumps(void*, '
                'unsigned long, nsAString_internal*, nsILocalFile**, '
                'nsILocalFile**)'
            ) in x
        ),
    ],
}


class BadRegularExpressionLineError(Exception):
    """Raised when a file contains an invalid regular expression."""


def _get_file_content(source):
    """Return a tuple, each value being a line of the source file.

    Remove empty lines and comments (lines starting with a '#').

    """
    filepath = os.path.join('siglists', source + '.txt')

    lines = []
    with resource_stream(__name__, filepath) as f:
        for i, line in enumerate(f):
            line = line.strip()
            if six.PY3:
                line = line.decode('utf-8', 'strict')

            if not line or line.startswith('#'):
                continue

            try:
                re.compile(line)
            except Exception as ex:
                raise BadRegularExpressionLineError(
                    'Regex error: {} in file {} at line {}'.format(
                        str(ex),
                        filepath,
                        i
                    )
                )

            lines.append(line)

    if source in _SPECIAL_EXTENDED_VALUES:
        lines = lines + _SPECIAL_EXTENDED_VALUES[source]

    return tuple(lines)


IRRELEVANT_SIGNATURE_RE = _get_file_content('irrelevant_signature_re')
PREFIX_SIGNATURE_RE = _get_file_content('prefix_signature_re')
SIGNATURE_SENTINELS = _get_file_content('signature_sentinels')
SIGNATURES_WITH_LINE_NUMBERS_RE = _get_file_content(
    'signatures_with_line_numbers_re'
)
TRIM_DLL_SIGNATURE_RE = _get_file_content('trim_dll_signature_re')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import argparse
import json
import sys

from .generator import SignatureGenerator


DESCRIPTION = """
Given a signature data structure as JSON via stdin, generates the signature.
"""


def main():
    """Takes crash data via stdin and generates a Socorro signature"""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-v', '--verbose', help='increase output verbosity', action='store_true'
    )
    args = parser.parse_args()

    generator = SignatureGenerator(debug=args.verbose)

    crash_data = json.loads(sys.stdin.read())

    ret = generator.generate(crash_data)

    print(json.dumps(ret, indent=2))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# NOTE(willkg): This is a duplicate of Socorro's WHATEVER, but it allows
# the signature generation code to stand alone.

from functools import total_ordering


@total_ordering
class EqualAnything(object):
    def __eq__(self, other):
        return True

    def __lt__(self, other):
        return True


#: Sentinel that is equal to anything; simplifies assertions in cases where
#: part of the value changes from test to test
WHATEVER = EqualAnything()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import importlib

import mock
from pkg_resources import resource_stream
import pytest

# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
base_module = '.'.join(__name__.split('.')[:-2])
siglists_utils = importlib.import_module(base_module + '.siglists_utils')


def _fake_stream(pkg, filepath):
    return resource_stream(__name__, filepath)


class TestSigLists:
    def test_loading_files(self):
        all_lists = (
            'IRRELEVANT_SIGNATURE_RE',
            'PREFIX_SIGNATURE_RE',
            'SIGNATURE_SENTINELS',
            'SIGNATURES_WITH_LINE_NUMBERS_RE',
        )

        for list_name in all_lists:
            content = getattr(siglists_utils, list_name)
            assert content

            for line in content:
                assert line
                if isinstance(line, basestring):
                    assert not line.startswith('#')

    @mock.patch(base_module + '.siglists_utils.resource_stream')
    def test_valid_entries(self, mocked_stream):
        mocked_stream.side_effect = _fake_stream

        expected = (
            'fooBarStuff',
            'moz::.*',
            '@0x[0-9a-fA-F]{2,}',
        )
        content = siglists_utils._get_file_content('test-valid-sig-list')
        assert content == expected

    @mock.patch(base_module + '.siglists_utils.resource_stream')
    def test_invalid_entry(self, mocked_stream):
        mocked_stream.side_effect = _fake_stream

        with pytest.raises(siglists_utils.BadRegularExpressionLineError) as exc_info:
            siglists_utils._get_file_content('test-invalid-sig-list')

        msg = exc_info.exconly()
        assert msg.startswith('BadRegularExpressionLineError: Regex error: ')
        assert msg.endswith('at line 3')
        assert 'test-invalid-sig-list.txt' in msg



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import importlib
import json
import re

import mock
import pytest

# NOTE(willkg): We do this so that we can extract signature generation into its
# own namespace as an external library. This allows the tests to run if it's in
# "siggen" or "socorro.signature".
base_module = '.'.join(__name__.split('.')[:-2])
rules = importlib.import_module(base_module + '.rules')


class TestCSignatureTool:

    @staticmethod
    def setup_config_c_sig_tool(
        ig=['ignored1'],
        pr=['pre1', 'pre2'],
        si=['fnNeedNumber'],
        td=['foo32\.dll.*'],
        ss=('sentinel', ('sentinel2', lambda x: 'ff' in x)),
    ):

        with mock.patch(base_module + '.rules.siglists_utils') as mocked_siglists:
            mocked_siglists.IRRELEVANT_SIGNATURE_RE = ig
            mocked_siglists.PREFIX_SIGNATURE_RE = pr
            mocked_siglists.SIGNATURES_WITH_LINE_NUMBERS_RE = si
            mocked_siglists.TRIM_DLL_SIGNATURE_RE = td
            mocked_siglists.SIGNATURE_SENTINELS = ss
            return rules.CSignatureTool()

    def test_c_config_tool_init(self):
        """test_C_config_tool_init: constructor test"""
        exp_irrelevant_signature_re = re.compile('ignored1')
        exp_prefix_signature_re = re.compile('pre1|pre2')
        exp_signatures_with_line_numbers_re = re.compile('fnNeedNumber')
        fixup_space = re.compile(r' (?=[\*&,])')
        fixup_comma = re.compile(r',(?! )')

        s = self.setup_config_c_sig_tool()

        assert exp_irrelevant_signature_re.pattern == s.irrelevant_signature_re.pattern
        assert exp_prefix_signature_re.pattern == s.prefix_signature_re.pattern
        assert exp_signatures_with_line_numbers_re.pattern == s.signatures_with_line_numbers_re.pattern  # noqa

        assert fixup_space.pattern == s.fixup_space.pattern
        assert fixup_comma.pattern == s.fixup_comma.pattern

    def test_normalize_with_collapse_args(self):
        """test_normalize: bunch of variations"""
        s = self.setup_config_c_sig_tool()
        a = [
            (('module', 'fn', 'source', '23', '0xFFF'), 'fn'),
            (('module', 'fnNeedNumber', 's', '23', '0xFFF'),
             'fnNeedNumber:23'),
            (('module', 'f( *s)', 's', '23', '0xFFF'), 'f'),
            (('module', 'f( &s)', 's', '23', '0xFFF'), 'f'),
            (('module', 'f( *s , &n)', 's', '23', '0xFFF'), 'f'),
            # this next one looks like a bug to me, but perhaps the situation
            # never comes up
            # (('module', 'f(  *s , &n)', 's', '23', '0xFFF'), 'f(*s, &n)'),
            (('module', 'f3(s,t,u)', 's', '23', '0xFFF'), 'f3'),
            (
                (
                    'module',
                    '::(anonymous namespace)::f3(s,t,u)',
                    's',
                    '23',
                    '0xFFF'
                ),
                '::(anonymous namespace)::f3'
            ),
            (
                (
                    'module',
                    'operator()(s,t,u)',
                    's',
                    '23',
                    '0xFFF'
                ),
                'operator()'
            ),
            (
                (
                    'module',
                    'Alpha<Bravo<Charlie>, Delta>::Echo<Foxtrot>',
                    's',
                    '23',
                    '0xFFF'
                ),
                'Alpha<T>::Echo<T>'
            ),
            (('module', 'f<3>(s,t,u)', 's', '23', '0xFFF'), 'f<T>'),
            (('module', '', 'source/', '23', '0xFFF'), 'source#23'),
            (('module', '', 'source\\', '23', '0xFFF'), 'source#23'),
            (('module', '', '/a/b/c/source', '23', '0xFFF'), 'source#23'),
            (('module', '', '\\a\\b\\c\\source', '23', '0xFFF'), 'source#23'),
            (('module', '', '\\a\\b\\c\\source', '23', '0xFFF'), 'source#23'),
            (('module', '', '\\a\\b\\c\\source', '', '0xFFF'), 'module@0xFFF'),
            (('module', '', '', '23', '0xFFF'), 'module@0xFFF'),
            (('module', '', '', '', '0xFFF'), 'module@0xFFF'),
            ((None, '', '', '', '0xFFF'), '@0xFFF'),
            (('module', 'expect_failed::h7f635057bfba806a', '', '', ''),
             'expect_failed'),
            (('module', 'expect_failed::h7f6350::blah', '', '', ''),
             'expect_failed::h7f6350::blah'),
        ]
        for args, e in a:
            r = s.normalize_signature(*args)
            assert e == r

    def test_generate_1(self):
        """test_generate_1: simple"""
        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghijklmnopqrstuvwxyz')
        sig, notes = s.generate(a)
        assert sig == 'd | e | f | g'

        a = list('abcdaeafagahijklmnopqrstuvwxyz')
        sig, notes = s.generate(a)
        assert sig == 'd | e | f | g'

    def test_generate_2(self):
        """test_generate_2: hang"""
        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghijklmnopqrstuvwxyz')
        sig, notes = s.generate(a, hang_type=-1)
        assert sig == 'hang | d | e | f | g'

        a = list('abcdaeafagahijklmnopqrstuvwxyz')
        sig, notes = s.generate(a, hang_type=-1)
        assert sig == 'hang | d | e | f | g'

        a = list('abcdaeafagahijklmnopqrstuvwxyz')
        sig, notes = s.generate(a, hang_type=0)
        assert sig == 'd | e | f | g'

        a = list('abcdaeafagahijklmnopqrstuvwxyz')
        sig, notes = s.generate(a, hang_type=1)
        assert sig == 'chromehang | d | e | f | g'

    def test_generate_2a(self):
        """test_generate_2a: way too long"""
        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghijklmnopqrstuvwxyz')
        a[3] = a[3] * 70
        a[4] = a[4] * 70
        a[5] = a[5] * 70
        a[6] = a[6] * 70
        a[7] = a[7] * 70
        expected = (
            'dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd | '
            'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee | '
            'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff | '
            'gggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggg'
        )
        sig, notes = s.generate(a)
        assert sig == expected
        expected = (
            'hang | '
            'dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd | '
            'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee | '
            'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff | '
            'gggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggg'
        )
        sig, notes = s.generate(a, hang_type=-1)
        assert sig == expected

    def test_generate_3(self):
        """test_generate_3: simple sentinel"""
        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghabcfaeabdijklmnopqrstuvwxyz')
        a[7] = 'sentinel'
        sig, notes = s.generate(a)
        assert sig == 'sentinel'

        s = self.setup_config_c_sig_tool(
            ['a', 'b', 'c', 'sentinel'],
            ['d', 'e', 'f']
        )
        sig, notes = s.generate(a)
        assert sig == 'f | e | d | i'

    def test_generate_4(self):
        """test_generate_4: tuple sentinel"""
        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghabcfaeabdijklmnopqrstuvwxyz')
        a[7] = 'sentinel2'
        sig, notes = s.generate(a)
        assert sig == 'd | e | f | g'

        s = self.setup_config_c_sig_tool(['a', 'b', 'c'], ['d', 'e', 'f'])
        a = list('abcdefghabcfaeabdijklmnopqrstuvwxyz')
        a[7] = 'sentinel2'
        a[22] = 'ff'
        sig, notes = s.generate(a)
        assert sig == 'sentinel2'

        s = self.setup_config_c_sig_tool(
            ['a', 'b', 'c', 'sentinel2'],
            ['d', 'e', 'f']
        )
        a = list('abcdefghabcfaeabdijklmnopqrstuvwxyz')
        a[7] = 'sentinel2'
        a[22] = 'ff'
        sig, notes = s.generate(a)
        assert sig == 'f | e | d | i'

    def test_specifiers_and_return_types(self):
        """prefix and irrelevant should match ignoring specifiers and return types"""
        generator = self.setup_config_c_sig_tool(
            # irrelevant
            ig=['ignored'],
            # prefix
            pr=['pre1', 'pre2'],
        )
        source_list = (
            'static void ignored',
            'pre1',
            'static bool pre2',
            'foo'
        )
        sig, notes = generator.generate(source_list)
        assert sig == 'pre1 | static bool pre2 | foo'

    def test_generate_with_merged_dll(self):
        generator = self.setup_config_c_sig_tool(
            ['a', 'b', 'c'],
            ['d', 'e', 'f']
        )
        source_list = (
            'a',
            'd',
            'foo32.dll@0x231423',
            'foo32.dll',
            'foo32.dll@0x42',
            'g',
        )
        sig, notes = generator.generate(source_list)
        assert sig == 'd | foo32.dll | g'

        source_list = (
            'foo32.dll',
            'foo32.dll@0x231423',
            'g',
        )
        sig, notes = generator.generate(source_list)
        assert sig == 'foo32.dll | g'


class TestJavaSignatureTool:
    def test_bad_stack(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = 17
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        assert sig == "EMPTY: Java stack trace not in expected format"
        assert notes == ['JavaSignatureTool: stack trace not in expected format']

    def test_basic_stack_frame_with_line_number(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            'SomeJavaException: totally made up  \n'
            'at org.mozilla.lars.myInvention(larsFile.java:666)'
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = 'SomeJavaException: totally made up at org.mozilla.lars.myInvention(larsFile.java)'
        assert sig == e
        assert notes == []

    def test_basic_stack_frame(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            'SomeJavaException: totally made up  \n'
            'at org.mozilla.lars.myInvention(larsFile.java)'
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = 'SomeJavaException: totally made up at org.mozilla.lars.myInvention(larsFile.java)'
        assert sig == e
        assert notes == []

    def test_long_exception_description(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            '   SomeJavaException: %s \nat org.mozilla.lars.myInvention(larsFile.java)' %
            ('t' * 1000)
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        expected = 'SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)'
        assert sig == expected
        expected = ['JavaSignatureTool: dropped Java exception description due to length']
        assert notes == expected

    def test_long_exception_description_with_line_number(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            '   SomeJavaException: %s  \nat org.mozilla.lars.myInvention(larsFile.java:1234)' %
            ('t' * 1000)
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        expected = 'SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)'
        assert sig == expected
        expected = ['JavaSignatureTool: dropped Java exception description due to length']
        assert notes == expected

    def test_no_description(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            '   SomeJavaException\n'
            'at org.mozilla.lars.myInvention(larsFile.java:1234)'
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = 'SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)'
        assert sig == e
        e = ['JavaSignatureTool: stack trace line 1 is not in the expected format']
        assert notes == e

    def test_frame_with_line_ending_but_missing_second_line(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = 'SomeJavaException: totally made up  \n'
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = 'SomeJavaException: totally made up'
        assert sig == e
        e = ['JavaSignatureTool: stack trace line 2 is missing']
        assert notes == e

    def test_frame_missing_second_line(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = 'SomeJavaException: totally made up  '
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = 'SomeJavaException: totally made up'
        assert sig == e
        e = ['JavaSignatureTool: stack trace line 2 is missing']
        assert notes == e

    def test_frame_with_leading_whitespace(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = (
            '   SomeJavaException: totally made up  \n'
            'at org.mozilla.lars.myInvention('
            'foolarsFile.java:1234)'
        )
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        expected = (
            'SomeJavaException: totally made up at org.mozilla.lars.myInvention(foolarsFile.java)'
        )
        assert sig == expected

    def test_no_interference(self):
        # In general addresses of the form ``@xxxxxxxx`` are to be replaced
        # with the literal ``<addr>``, however in this case, the hex address is
        # not in the expected location and should therefore be left alone
        j = rules.JavaSignatureTool()
        java_stack_trace = ('SomeJavaException: totally made up  \n'
                            'at org.mozilla.lars.myInvention('
                            'larsFile.java:@abef1234)')
        sig, notes = j.generate(java_stack_trace, delimiter=' ')
        e = ('SomeJavaException totally made up '
             'at org.mozilla.lars.myInvention('
             'larsFile.java:@abef1234)')
        assert sig == e
        assert notes == []

    def test_replace_address(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
java.lang.IllegalArgumentException: Given view not a child of android.widget.AbsoluteLayout@4054b560
\tat android.view.ViewGroup.updateViewLayout(ViewGroup.java:1968)
\tat org.mozilla.gecko.GeckoApp.repositionPluginViews(GeckoApp.java:1492)
\tat org.mozilla.gecko.GeckoApp.repositionPluginViews(GeckoApp.java:1475)
\tat org.mozilla.gecko.gfx.LayerController$2.run(LayerController.java:269)
\tat android.os.Handler.handleCallback(Handler.java:587)
\tat android.os.Handler.dispatchMessage(Handler.java:92)
\tat android.os.Looper.loop(Looper.java:150)
\tat org.mozilla.gecko.GeckoApp$32.run(GeckoApp.java:1670)
\tat android.os.Handler.handleCallback(Handler.java:587)
\tat android.os.Handler.dispatchMessage(Handler.java:92)
\tat android.os.Looper.loop(Looper.java:150)
\tat android.app.ActivityThread.main(ActivityThread.java:4293)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:849)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:607)
\tat dalvik.system.NativeStart.main(Native Method)""".lstrip()
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = (
            'java.lang.IllegalArgumentException: '
            'Given view not a child of android.widget.AbsoluteLayout@<addr>: '
            'at android.view.ViewGroup.updateViewLayout(ViewGroup.java)'
        )
        assert sig == e
        assert notes == []

    def test_replace_address_with_trailing_text(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
android.view.WindowManager$BadTokenException: Unable to add window -- token android.os.BinderProxy@406237c0 is not valid; is your activity running?
\tat android.view.ViewRoot.setView(ViewRoot.java:533)
\tat android.view.WindowManagerImpl.addView(WindowManagerImpl.java:202)
\tat android.view.WindowManagerImpl.addView(WindowManagerImpl.java:116)
\tat android.view.Window$LocalWindowManager.addView(Window.java:424)
\tat android.app.ActivityThread.handleResumeActivity(ActivityThread.java:2174)
\tat android.app.ActivityThread.handleLaunchActivity(ActivityThread.java:1672)
\tat android.app.ActivityThread.access$1500(ActivityThread.java:117)
\tat android.app.ActivityThread$H.handleMessage(ActivityThread.java:935)
\tat android.os.Handler.dispatchMessage(Handler.java:99)
\tat android.os.Looper.loop(Looper.java:130)
\tat android.app.ActivityThread.main(ActivityThread.java:3687)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:867)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:625)
\tat dalvik.system.NativeStart.main(Native Method)
""".lstrip()  # noqa
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = ('android.view.WindowManager$BadTokenException: '
             'Unable to add window -- token android.os.BinderProxy@<addr> '
             'is not valid; is your activity running? '
             'at android.view.ViewRoot.setView(ViewRoot.java)')
        assert sig == e
        assert notes == []

    def test_replace_address_trailing_whitespace(self):
        j = rules.JavaSignatureTool()
        java_stack_trace = """
java.lang.IllegalArgumentException: Receiver not registered: org.mozilla.gecko.GeckoNetworkManager@405afea8
\tat android.app.LoadedApk.forgetReceiverDispatcher(LoadedApk.java:610)
\tat android.app.ContextImpl.unregisterReceiver(ContextImpl.java:883)
\tat android.content.ContextWrapper.unregisterReceiver(ContextWrapper.java:331)
\tat org.mozilla.gecko.GeckoNetworkManager.stopListening(GeckoNetworkManager.java:141)
\tat org.mozilla.gecko.GeckoNetworkManager.stop(GeckoNetworkManager.java:136)
\tat org.mozilla.gecko.GeckoApp.onApplicationPause(GeckoApp.java:2130)
\tat org.mozilla.gecko.GeckoApplication.onActivityPause(GeckoApplication.java:55)
\tat org.mozilla.gecko.GeckoActivity.onPause(GeckoActivity.java:22)
\tat org.mozilla.gecko.GeckoApp.onPause(GeckoApp.java:1948)
\tat android.app.Activity.performPause(Activity.java:3877)
\tat android.app.Instrumentation.callActivityOnPause(Instrumentation.java:1191)
\tat android.app.ActivityThread.performPauseActivity(ActivityThread.java:2345)
\tat android.app.ActivityThread.performPauseActivity(ActivityThread.java:2315)
\tat android.app.ActivityThread.handlePauseActivity(ActivityThread.java:2295)
\tat android.app.ActivityThread.access$1700(ActivityThread.java:117)
\tat android.app.ActivityThread$H.handleMessage(ActivityThread.java:942)
\tat android.os.Handler.dispatchMessage(Handler.java:99)
\tat android.os.Looper.loop(Looper.java:130)
\tat android.app.ActivityThread.main(ActivityThread.java:3691)
\tat java.lang.reflect.Method.invokeNative(Native Method)
\tat java.lang.reflect.Method.invoke(Method.java:507)
\tat com.android.internal.os.ZygoteInit$MethodAndArgsCaller.run(ZygoteInit.java:907)
\tat com.android.internal.os.ZygoteInit.main(ZygoteInit.java:665)
\tat dalvik.system.NativeStart.main(Native Method)
""".lstrip()  # noqa
        sig, notes = j.generate(java_stack_trace, delimiter=': ')
        e = (
            'java.lang.IllegalArgumentException: '
            'Receiver not registered: '
            'org.mozilla.gecko.GeckoNetworkManager@<addr>: '
            'at android.app.LoadedApk.forgetReceiverDispatcher(LoadedApk.java)'
        )
        assert sig == e
        assert notes == []


#  rules testing section

frames_from_json_dump = {
    'frames': [
        {
            'frame': 0,
            'function': 'NtWaitForMultipleObjects',
            'function_offset': '0x15',
            'module': 'ntdll.dll',
            'module_offset': '0x2015d',
            'offset': '0x77ad015d',
            'registers': {
                'eax': '0x00000040',
                'ebp': '0x0025e968',
                'ebx': '0x0025e91c',
                'ecx': '0x00000000',
                'edi': '0x00000000',
                'edx': '0x00000000',
                'efl': '0x00200246',
                'eip': '0x77ad015d',
                'esi': '0x00000004',
                'esp': '0x0025e8cc'
            },
            'trust': 'context'
        },
        {
            'frame': 1,
            'function': 'WaitForMultipleObjectsEx',
            'function_offset': '0xff',
            'module': 'KERNELBASE.dll',
            'module_offset': '0x115f6',
            'offset': '0x775e15f6',
            'trust': 'cfi'
        },
        {
            'frame': 2,
            'function': 'WaitForMultipleObjectsExImplementation',
            'function_offset': '0x8d',
            'module': 'kernel32.dll',
            'module_offset': '0x119f7',
            'offset': '0x766119f7',
            'trust': 'cfi'
        },
        {
            'frame': 3,
            'function': 'RealMsgWaitForMultipleObjectsEx',
            'function_offset': '0xe1',
            'module': 'user32.dll',
            'module_offset': '0x20869',
            'offset': '0x77370869',
            'trust': 'cfi'
        },
        {
            'frame': 4,
            'function': 'MsgWaitForMultipleObjects',
            'function_offset': '0x1e',
            'module': 'user32.dll',
            'module_offset': '0x20b68',
            'offset': '0x77370b68',
            'trust': 'cfi'
        },
        {
            'file': 'F117835525________________________________________',
            'frame': 5,
            'function': 'F_1152915508__________________________________',
            'function_offset': '0xbb',
            'line': 118,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x36a13b',
            'offset': '0x5e3aa13b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 6,
            'function': 'F2166389______________________________________',
            'function_offset': '0xe5',
            'line': 552,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35faf5',
            'offset': '0x5e39faf5',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 7,
            'function': 'F_917831355___________________________________',
            'function_offset': '0x29b',
            'line': 488,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x360a7b',
            'offset': '0x5e3a0a7b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 8,
            'function': 'F1315696776________________________________',
            'function_offset': '0xd',
            'line': 439,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35e2fd',
            'offset': '0x5e39e2fd',
            'trust': 'cfi'
        },
        {
            'file': 'F_766591945_______________________________________',
            'frame': 9,
            'function': 'F_1428703866________________________________',
            'function_offset': '0xc1',
            'line': 203,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35bf21',
            'offset': '0x5e39bf21',
            'trust': 'cfi'
        }
    ],
    'threads_index': 0,
    'frame_count': 32
}

frames_from_json_dump_with_templates = {
    'frames': [
        {
            'frame': 0,
            'function': 'NtWaitForMultipleObjects',
            'function_offset': '0x15',
            'module': 'ntdll.dll',
            'module_offset': '0x2015d',
            'offset': '0x77ad015d',
            'registers': {
                'eax': '0x00000040',
                'ebp': '0x0025e968',
                'ebx': '0x0025e91c',
                'ecx': '0x00000000',
                'edi': '0x00000000',
                'edx': '0x00000000',
                'efl': '0x00200246',
                'eip': '0x77ad015d',
                'esi': '0x00000004',
                'esp': '0x0025e8cc'
            },
            'trust': 'context'
        },
        {
            'frame': 1,
            'function': 'Alpha<Bravo<Charlie>, Delta>::Echo<Foxtrot>',
            'function_offset': '0xff',
            'module': 'KERNELBASE.dll',
            'module_offset': '0x115f6',
            'offset': '0x775e15f6',
            'trust': 'cfi'
        },
        {
            'frame': 2,
            'function': 'WaitForMultipleObjectsExImplementation',
            'function_offset': '0x8d',
            'module': 'kernel32.dll',
            'module_offset': '0x119f7',
            'offset': '0x766119f7',
            'trust': 'cfi'
        },
        {
            'frame': 3,
            'function': 'RealMsgWaitForMultipleObjectsEx(void **fakeargs)',
            'function_offset': '0xe1',
            'module': 'user32.dll',
            'module_offset': '0x20869',
            'offset': '0x77370869',
            'trust': 'cfi'
        },
        {
            'frame': 4,
            'function': 'MsgWaitForMultipleObjects',
            'function_offset': '0x1e',
            'module': 'user32.dll',
            'module_offset': '0x20b68',
            'offset': '0x77370b68',
            'trust': 'cfi'
        },
        {
            'file': 'F117835525________________________________________',
            'frame': 5,
            'function': 'F_1152915508__________________________________',
            'function_offset': '0xbb',
            'line': 118,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x36a13b',
            'offset': '0x5e3aa13b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 6,
            'function': 'F2166389______________________________________',
            'function_offset': '0xe5',
            'line': 552,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35faf5',
            'offset': '0x5e39faf5',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 7,
            'function': 'F_917831355___________________________________',
            'function_offset': '0x29b',
            'line': 488,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x360a7b',
            'offset': '0x5e3a0a7b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 8,
            'function': 'F1315696776________________________________',
            'function_offset': '0xd',
            'line': 439,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35e2fd',
            'offset': '0x5e39e2fd',
            'trust': 'cfi'
        },
        {
            'file': 'F_766591945_______________________________________',
            'frame': 9,
            'function': 'F_1428703866________________________________',
            'function_offset': '0xc1',
            'line': 203,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35bf21',
            'offset': '0x5e39bf21',
            'trust': 'cfi'
        }
    ],
    'threads_index': 0,
    'total_frames': 32
}

frames_from_json_dump_with_templates_and_special_case = {
    'frames': [
        {
            'frame': 0,
            'function': 'NtWaitForMultipleObjects',
            'function_offset': '0x15',
            'module': 'ntdll.dll',
            'module_offset': '0x2015d',
            'offset': '0x77ad015d',
            'registers': {
                'eax': '0x00000040',
                'ebp': '0x0025e968',
                'ebx': '0x0025e91c',
                'ecx': '0x00000000',
                'edi': '0x00000000',
                'edx': '0x00000000',
                'efl': '0x00200246',
                'eip': '0x77ad015d',
                'esi': '0x00000004',
                'esp': '0x0025e8cc'
            },
            'trust': 'context'
        },
        {
            'frame': 1,
            'function': '<name omitted>',
            'function_offset': '0xff',
            'module': 'KERNELBASE.dll',
            'module_offset': '0x115f6',
            'offset': '0x775e15f6',
            'trust': 'cfi'
        },
        {
            'frame': 2,
            'function': 'IPC::ParamTraits<mozilla::net::NetAddr>::Write',
            'function_offset': '0x8d',
            'module': 'kernel32.dll',
            'module_offset': '0x119f7',
            'offset': '0x766119f7',
            'trust': 'cfi'
        },
        {
            'frame': 3,
            'function': 'RealMsgWaitForMultipleObjectsEx(void **fakeargs)',
            'function_offset': '0xe1',
            'module': 'user32.dll',
            'module_offset': '0x20869',
            'offset': '0x77370869',
            'trust': 'cfi'
        },
        {
            'frame': 4,
            'function': 'MsgWaitForMultipleObjects',
            'function_offset': '0x1e',
            'module': 'user32.dll',
            'module_offset': '0x20b68',
            'offset': '0x77370b68',
            'trust': 'cfi'
        },
        {
            'file': 'F117835525________________________________________',
            'frame': 5,
            'function': 'F_1152915508__________________________________',
            'function_offset': '0xbb',
            'line': 118,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x36a13b',
            'offset': '0x5e3aa13b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 6,
            'function': 'F2166389______________________________________',
            'function_offset': '0xe5',
            'line': 552,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35faf5',
            'offset': '0x5e39faf5',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 7,
            'function': 'F_917831355___________________________________',
            'function_offset': '0x29b',
            'line': 488,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x360a7b',
            'offset': '0x5e3a0a7b',
            'trust': 'cfi'
        },
        {
            'file': 'F_851861807_______________________________________',
            'frame': 8,
            'function': 'F1315696776________________________________',
            'function_offset': '0xd',
            'line': 439,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35e2fd',
            'offset': '0x5e39e2fd',
            'trust': 'cfi'
        },
        {
            'file': 'F_766591945_______________________________________',
            'frame': 9,
            'function': 'F_1428703866________________________________',
            'function_offset': '0xc1',
            'line': 203,
            'module': 'NPSWF32_14_0_0_125.dll',
            'module_offset': '0x35bf21',
            'offset': '0x5e39bf21',
            'trust': 'cfi'
        }
    ],
    'threads_index': 0,
    'total_frames': 32
}


class TestSignatureGeneration:

    def test_create_frame_list(self):
        sgr = rules.SignatureGenerationRule()
        frame_signatures_list = sgr._create_frame_list(frames_from_json_dump)
        expected = [
            'NtWaitForMultipleObjects',
            'WaitForMultipleObjectsEx',
            'WaitForMultipleObjectsExImplementation',
            'RealMsgWaitForMultipleObjectsEx',
            'MsgWaitForMultipleObjects',
            'F_1152915508__________________________________',
            'F2166389______________________________________',
            'F_917831355___________________________________',
            'F1315696776________________________________',
            'F_1428703866________________________________'
        ]
        assert frame_signatures_list == expected
        assert 'normalized' in frames_from_json_dump['frames'][0]
        assert frames_from_json_dump['frames'][0]['normalized'] == expected[0]

    def test_java_stack_trace(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'java_stack_trace': (
                '   SomeJavaException: %s  \nat org.mozilla.lars.myInvention(larsFile.java)' %
                ('t' * 1000)
            )
        }

        signature = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, signature) is True

        expected = 'SomeJavaException: at org.mozilla.lars.myInvention(larsFile.java)'
        assert signature['signature'] == expected
        assert 'proto_signature' not in signature
        expected = ['JavaSignatureTool: dropped Java exception description due to length']
        assert signature['notes'] == expected

    def test_c_stack_trace(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'os': 'Windows NT',
            'threads': [frames_from_json_dump]
        }
        result = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        expected = 'MsgWaitForMultipleObjects | F_1152915508__________________________________'
        assert result['signature'] == expected

        expected = (
            'NtWaitForMultipleObjects | WaitForMultipleObjectsEx | '
            'WaitForMultipleObjectsExImplementation | '
            'RealMsgWaitForMultipleObjectsEx | MsgWaitForMultipleObjects | '
            'F_1152915508__________________________________ | '
            'F2166389______________________________________ | '
            'F_917831355___________________________________ | '
            'F1315696776________________________________ | '
            'F_1428703866________________________________'
        )
        assert result['proto_signature'] == expected
        assert result['notes'] == []

    def test_action_2_with_templates(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'os': 'Windows NT',
            'crashing_thread': 0,
            'threads': [frames_from_json_dump_with_templates]
        }
        result = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        assert result['signature'] == 'Alpha<T>::Echo<T>'
        expected = (
            'NtWaitForMultipleObjects | Alpha<T>::Echo<T> | '
            'WaitForMultipleObjectsExImplementation | '
            'RealMsgWaitForMultipleObjectsEx | '
            'MsgWaitForMultipleObjects | '
            'F_1152915508__________________________________ | '
            'F2166389______________________________________ | '
            'F_917831355___________________________________ | '
            'F1315696776________________________________ | '
            'F_1428703866________________________________'
        )
        assert result['proto_signature'] == expected
        assert result['notes'] == []

    def test_action_2_with_templates_and_special_case(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'os': 'Windows NT',
            'crashing_thread': 0,
            'threads': [frames_from_json_dump_with_templates_and_special_case]
        }
        result = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        expected = '<name omitted> | IPC::ParamTraits<mozilla::net::NetAddr>::Write'
        assert result['signature'] == expected
        expected = (
            'NtWaitForMultipleObjects | '
            '<name omitted> | '
            'IPC::ParamTraits<mozilla::net::NetAddr>::Write | '
            'RealMsgWaitForMultipleObjectsEx | '
            'MsgWaitForMultipleObjects | '
            'F_1152915508__________________________________ | '
            'F2166389______________________________________ | '
            'F_917831355___________________________________ | '
            'F1315696776________________________________ | '
            'F_1428703866________________________________'
        )
        assert result['proto_signature'] == expected
        assert result['notes'] == []

    def test_action_3(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'thread': [[]],
        }
        result = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        assert result['signature'] == 'EMPTY: no crashing thread identified'
        assert 'proto_signature' not in result
        expected = [
            'CSignatureTool: No signature could be created because we do '
            'not know which thread crashed'
        ]
        assert result['notes'] == expected

    def test_lower_case_modules(self):
        sgr = rules.SignatureGenerationRule()

        crash_data = {
            'os': 'Windows NT',
            'threads': [{
                "frames": [
                    {
                        'offset': '0x5e39bf21',
                        'trust': 'cfi'
                    },
                    {
                        'offset': '0x5e39bf21',
                        'trust': 'cfi'
                    },
                    {
                        'offset': '0x5e39bf21',
                        'trust': 'cfi'
                    },
                    {
                        'frame': 3,
                        'module': 'USER2.dll',
                        'module_offset': '0x20869',
                        'offset': '0x77370869',
                        'trust': 'cfi'
                    },
                ]
            }]
        }
        result = {
            'signature': '',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True
        assert result['signature'] == 'user2.dll@0x20869'
        expected = '@0x5e39bf21 | @0x5e39bf21 | @0x5e39bf21 | user2.dll@0x20869'
        assert result['proto_signature'] == expected
        assert result['notes'] == []


class TestOOMSignature:
    def test_predicate_no_match(self):
        result = {
            'signature': 'hello',
            'notes': []
        }
        rule = rules.OOMSignature()
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        crash_data = {
            'oom_allocation_size': 17
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_1(self):
        crash_data = {}
        result = {
            'signature': 'this | is | a | NS_ABORT_OOM | signature',
            'notes': []
        }
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_2(self):
        crash_data = {}
        result = {
            'signature': 'mozalloc_handle_oom | this | is | bad',
            'notes': []
        }
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_predicate_signature_fragment_3(self):
        crash_data = {}
        result = {
            'signature': 'CrashAtUnhandlableOOM',
            'notes': []
        }
        rule = rules.OOMSignature()
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        crash_data = {}
        result = {
            'signature': 'hello',
            'notes': []
        }
        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result['signature'] == 'OOM | unknown | hello'

    def test_action_small(self):
        crash_data = {
            'oom_allocation_size': 17
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result['signature'] == 'OOM | small'

    def test_action_large(self):
        crash_data = {
            'oom_allocation_size': 17000000
        }
        result = {
            'signature': 'hello',
            'notes': []
        }

        rule = rules.OOMSignature()
        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result['signature'] == 'OOM | large | hello'


class TestAbortSignature:

    def test_predicate(self):
        rule = rules.AbortSignature()
        crash_data = {
            'abort_message': 'something'
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is True

    def test_predicate_no_match(self):
        rule = rules.AbortSignature()
        # No AbortMessage
        crash_data = {}
        result = {
            'signature': 'hello'
        }
        assert rule.predicate(crash_data, result) is False

    def test_predicate_empty_message(self):
        rule = rules.AbortSignature()
        crash_data = {
            'abort_message': ''
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is False

    def test_action_success(self):
        rule = rules.AbortSignature()
        crash_data = {
            'abort_message': 'unknown'
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'Abort | unknown | hello'

    def test_action_success_long_message(self):
        rule = rules.AbortSignature()

        crash_data = {
            'abort_message': 'a' * 81
        }
        result = {
            'signature': 'hello',
            'notes': []
        }

        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result['signature'] == 'Abort | {}... | hello'.format('a' * 77)

    @pytest.mark.parametrize('abort_msg, expected', [
        # Test with just the "ABOR" thing at the start
        (
            '[5392] ###!!! ABORT: foo bar line 42',
            'Abort | foo bar line 42 | hello'
        ),

        # Test with a file name and line number
        (
            (
                '[7616] ###!!! ABORT: unsafe destruction: file '
                'c:/builds/moz2_slave/m-rel-w32-00000000000000000000/build/src/'
                'dom/plugins/ipc/PluginModuleParent.cpp, line 777'
            ),
            'Abort | unsafe destruction | hello'
        ),

        # Test with a message that lacks interesting content.
        (
            '[204] ###!!! ABORT: file ?, ',
            'Abort | hello'
        ),

        # Test with another message that lacks interesting content.
        (
            (
                '[4648] ###!!! ABORT: file resource:///modules/sessionstore/'
                'SessionStore.jsm, line 1459'
            ),
            'Abort | hello'
        ),

        # Test with "unable to find a usable font" case
        (
            'unable to find a usable font (\u5fae\u8f6f\u96c5\u9ed1)',
            'Abort | unable to find a usable font | hello'
        ),
    ])
    def test_action_success_remove_unwanted_parts(self, abort_msg, expected):
        rule = rules.AbortSignature()

        crash_data = {
            'abort_message': abort_msg
        }
        result = {
            'signature': 'hello'
        }

        action_result = rule.action(crash_data, result)

        assert action_result is True
        assert result['signature'] == expected

    def test_action_non_ascii_abort_message(self):
        # Non-ascii characters are removed from abort messages
        rule = rules.AbortSignature()
        crash_data = {
            'abort_message': '\u018a unknown'
        }
        result = {
            'signature': 'hello',
            'notes': []
        }
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'Abort | unknown | hello'


class TestSigFixWhitespace:

    def test_predicate_no_match(self):
        rule = rules.SigFixWhitespace()

        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate({}, result) is True

        result['signature'] = 42
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.SigFixWhitespace()
        result = {
            'signature': 'fooo::baar',
            'notes': []
        }
        assert rule.predicate({}, result) is True

    @pytest.mark.parametrize('signature, expected', [
        # Leading and trailing whitespace are removed
        ('all   good', 'all good'),
        ('all   good     ', 'all good'),
        ('    all   good  ', 'all good'),

        # Non-space whitespace is converted to spaces
        ('all\tgood', 'all good'),
        ('all\n\ngood', 'all good'),

        # Multiple consecutive spaces are converted to a single space
        ('all   good', 'all good'),
        ('all  |  good', 'all | good'),
    ])
    def test_whitespace_fixing(self, signature, expected):
        rule = rules.SigFixWhitespace()
        result = {
            'signature': signature,
            'notes': []
        }
        action_result = rule.action({}, result)
        assert action_result is True
        assert result['signature'] == expected


class TestSigTruncate:

    def test_predicate_no_match(self):
        rule = rules.SigTruncate()
        result = {
            'signature': '0' * 100,
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.SigTruncate()
        result = {
            'signature': '9' * 256,
            'notes': []
        }
        assert rule.predicate({}, result) is True

    def test_action_success(self):
        rule = rules.SigTruncate()
        result = {
            'signature': '9' * 256,
            'notes': []
        }
        action_result = rule.action({}, result)
        assert action_result is True
        assert len(result['signature']) == 255
        assert result['signature'].endswith('9...')


class TestStackwalkerErrorSignatureRule:

    def test_predicate_no_match_signature(self):
        rule = rules.StackwalkerErrorSignatureRule()
        result = {
            'signature': '0' * 100,
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate_no_match_missing_mdsw_status_string(self):
        rule = rules.StackwalkerErrorSignatureRule()
        result = {
            'signature': 'EMPTY: like my soul',
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.StackwalkerErrorSignatureRule()
        crash_data = {
            'mdsw_status_string': 'catastrophic stackwalker failure'
        }
        result = {
            'signature': 'EMPTY: like my soul',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.StackwalkerErrorSignatureRule()
        crash_data = {
            'mdsw_status_string': 'catastrophic stackwalker failure'
        }
        result = {
            'signature': 'EMPTY: like my soul',
            'notes': []
        }
        action_result = rule.action(crash_data, result)
        assert action_result is True
        expected = 'EMPTY: like my soul; catastrophic stackwalker failure'
        assert result['signature'] == expected


class TestSignatureWatchDogRule:

    def test_instantiation(self):
        srwd = rules.SignatureRunWatchDog()

        assert isinstance(srwd.c_signature_tool, rules.CSignatureTool)
        assert isinstance(srwd.java_signature_tool, rules.JavaSignatureTool)

        assert srwd._get_crashing_thread({}) == 0

    def test_predicate(self):
        srwd = rules.SignatureRunWatchDog()

        result = {
            'signature': "I'm not real",
            'notes': []
        }
        assert srwd.predicate({}, result) is False

        result = {
            'signature': "mozilla::`anonymous namespace''::RunWatchdog(void*)",
            'notes': []
        }
        assert srwd.predicate({}, result) is True

        result = {
            'signature': "mozilla::(anonymous namespace)::RunWatchdog",
            'notes': []
        }
        assert srwd.predicate({}, result) is True

    def test_action(self):
        sgr = rules.SignatureRunWatchDog()

        crash_data = {
            'os': 'Windows NT',
            'crashing_thread': 0,
            'threads': [frames_from_json_dump]
        }
        result = {
            'signature': 'foo::bar',
            'notes': []
        }

        # the call to be tested
        assert sgr.action(crash_data, result) is True

        # Verify the signature has been re-generated based on thread 0.
        expected = (
            'shutdownhang | MsgWaitForMultipleObjects | '
            'F_1152915508__________________________________'
        )
        assert result['signature'] == expected
        assert result['notes'] == []


class TestSignatureJitCategory:

    def test_predicate_no_match(self):
        rule = rules.SignatureJitCategory()

        crash_data = {}
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is False

        crash_data = {
            'jit_category': ''
        }
        result = {
            'signature': '',
            'notes': []
        }

        assert rule.predicate(crash_data, result) is False

    def test_predicate(self):
        rule = rules.SignatureJitCategory()

        crash_data = {
            'jit_category': 'JIT Crash'
        }
        result = {
            'signature': '',
            'notes': []
        }

        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.SignatureJitCategory()

        crash_data = {
            'jit_category': 'JIT Crash'
        }
        result = {
            'signature': 'foo::bar',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'jit | JIT Crash'
        assert (
            result['notes'] ==
            ['Signature replaced with a JIT Crash Category, was: "foo::bar"']
        )


class TestSignatureIPCChannelError:

    def test_predicate_no_match(self):
        rule = rules.SignatureIPCChannelError()

        result = {
            'signature': '',
            'notes': []
        }

        assert rule.predicate({}, result) is False

        crash_data = {
            'ipc_channel_error': ''
        }
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is False

    def test_predicate(self):
        rule = rules.SignatureIPCChannelError()

        crash_data = {
            'ipc_channel_error': 'foo, bar'
        }
        result = {
            'signature': '',
            'notes': []
        }

        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.SignatureIPCChannelError()

        crash_data = {
            'ipc_channel_error': 'ipc' * 50
        }
        result = {
            'signature': 'foo::bar',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        expected = 'IPCError-content | {}'.format(('ipc' * 50)[:100])
        assert result['signature'] == expected
        assert (
            result['notes'] ==
            ['Signature replaced with an IPC Channel Error, was: "foo::bar"']
        )

        # Now test with a browser crash.
        crash_data['additional_minidumps'] = 'browser'
        result = {
            'signature': 'foo::bar',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True

        assert result['signature'] == 'IPCError-browser | {}'.format(('ipc' * 50)[:100])
        assert (
            result['notes'] ==
            ['Signature replaced with an IPC Channel Error, was: "foo::bar"']
        )


class TestSignatureShutdownTimeout:

    def test_predicate_no_match(self):
        rule = rules.SignatureShutdownTimeout()
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': '{"foo": "bar"}'
        }
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is True

    def test_action_missing_valueerror(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': '{{{{'
        }
        result = {
            'signature': 'foo',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'AsyncShutdownTimeout | UNKNOWN'

        assert 'Error parsing AsyncShutdownTimeout:' in result['notes'][0]
        assert 'Expected object or value' in result['notes'][0]
        assert (
            'Signature replaced with a Shutdown Timeout signature, was: "foo"' in result['notes'][1]
        )

    def test_action_missing_keyerror(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': json.dumps({'no': 'phase or condition'})
        }
        result = {
            'signature': 'foo',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'AsyncShutdownTimeout | UNKNOWN'

        assert result['notes'][0] == "Error parsing AsyncShutdownTimeout: 'phase'"
        assert (
            result['notes'][1] ==
            'Signature replaced with a Shutdown Timeout signature, was: "foo"'
        )

    def test_action_success(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': json.dumps({
                'phase': 'beginning',
                'conditions': [
                    {'name': 'A'},
                    {'name': 'B'},
                ]
            })
        }
        result = {
            'signature': 'foo',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'AsyncShutdownTimeout | beginning | A,B'
        expected = 'Signature replaced with a Shutdown Timeout signature, was: "foo"'
        assert result['notes'][0] == expected

    def test_action_success_string_conditions(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': json.dumps({
                'phase': 'beginning',
                'conditions': ['A', 'B', 'C']
            })
        }
        result = {
            'signature': 'foo',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'AsyncShutdownTimeout | beginning | A,B,C'
        expected = 'Signature replaced with a Shutdown Timeout signature, was: "foo"'
        assert result['notes'][0] == expected

    def test_action_success_empty_conditions_key(self):
        rule = rules.SignatureShutdownTimeout()

        crash_data = {
            'async_shutdown_timeout': json.dumps({
                'phase': 'beginning',
                'conditions': []
            })
        }
        result = {
            'signature': 'foo',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'AsyncShutdownTimeout | beginning | (none)'
        expected = 'Signature replaced with a Shutdown Timeout signature, was: "foo"'
        assert result['notes'][0] == expected


class TestSignatureIPCMessageName:

    def test_predicate_no_ipc_message_name(self):
        rule = rules.SignatureIPCMessageName()
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate_empty_string(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {
            'ipc_message_name': ''
        }
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is False

    def test_predicate(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {
            'ipc_message_name': 'foo, bar'
        }
        result = {
            'signature': 'fooo::baar',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is True

    def test_action_success(self):
        rule = rules.SignatureIPCMessageName()
        crash_data = {
            'ipc_message_name': 'foo, bar'
        }
        result = {
            'signature': 'fooo::baar',
            'notes': []
        }
        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'fooo::baar | IPC_Message_Name=foo, bar'


class TestSignatureParentIDNotEqualsChildID:

    def test_predicate_no_moz_crash_reason(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate({}, result) is False

    def test_predicate_empty_moz_crash_reason(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {
            'moz_crash_reason': ''
        }
        result = {
            'signature': '',
            'notes': []
        }
        assert rule.predicate(crash_data, result) is False

    def test_predicate_match(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {
            'moz_crash_reason': 'MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)'
        }
        result = {
            'signature': 'fooo::baar',
            'notes': []
        }

        assert rule.predicate(crash_data, result) is True

    def test_action(self):
        rule = rules.SignatureParentIDNotEqualsChildID()
        crash_data = {
            'moz_crash_reason': 'MOZ_RELEASE_ASSERT(parentBuildID == childBuildID)'
        }
        result = {
            'signature': 'fooo::baar',
            'notes': []
        }

        action_result = rule.action(crash_data, result)
        assert action_result is True
        assert result['signature'] == 'parentBuildID != childBuildID'
        assert result['notes'][0] == 'Signature replaced with MozCrashAssert, was: "fooo::baar"'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import argparse
import json
import os
import sys

import requests

from .utils import convert_to_crash_data


class WrappedTextHelpFormatter(argparse.HelpFormatter):
    """Subclass that wraps description and epilog text taking paragraphs into account"""

    def _fill_text(self, text, width, indent):
        """Wraps text like HelpFormatter, but doesn't squash lines

        This makes it easier to do lists and paragraphs.

        """
        parts = text.split('\n\n')
        for i, part in enumerate(parts):
            # Check to see if it's a bulleted list--if so, then fill each line
            if part.startswith('* '):
                subparts = part.split('\n')
                for j, subpart in enumerate(subparts):
                    subparts[j] = super(WrappedTextHelpFormatter, self)._fill_text(
                        subpart, width, indent
                    )
                parts[i] = '\n'.join(subparts)
            else:
                parts[i] = super(WrappedTextHelpFormatter, self)._fill_text(part, width, indent)

        return '\n\n'.join(parts)


DESCRIPTION = """
Takes a crash id via the command line, pulls down the crash information, and
outputs JSON for signature generation.

To use an API token from Crash Stats to alleviate issues with rate limiting,
set SOCORRO_API_TOKEN in the environment.
"""

# FIXME(willkg): This hits production. We might want it configurable.
API_URL = 'https://crash-stats.mozilla.com/api'


def printerr(s, **kwargs):
    kwargs['file'] = sys.stderr
    print(s, **kwargs)


def fetch(endpoint, crash_id, api_token=None):
    kwargs = {
        'params': {
            'crash_id': crash_id
        }
    }
    if api_token:
        kwargs['headers'] = {
            'Auth-Token': api_token
        }

    return requests.get(API_URL + endpoint, **kwargs)


def main():
    """Takes a crash id, pulls down data from Socorro, generates signature data"""
    parser = argparse.ArgumentParser(
        formatter_class=WrappedTextHelpFormatter,
        description=DESCRIPTION
    )
    parser.add_argument(
        '-v', '--verbose', help='increase output verbosity', action='store_true'
    )
    parser.add_argument(
        'crashid', help='crash id to generate signatures for'
    )

    args = parser.parse_args()

    api_token = os.environ.get('SOCORRO_API_TOKEN', '')

    crash_id = args.crashid.strip()

    resp = fetch('/RawCrash/', crash_id, api_token)
    if resp.status_code == 404:
        printerr('%s: does not exist.' % crash_id)
        return 1
    if resp.status_code == 429:
        printerr('API rate limit reached. %s' % resp.content)
        # FIXME(willkg): Maybe there's something better we could do here. Like maybe wait a
        # few minutes.
        return 1
    if resp.status_code == 500:
        printerr('HTTP 500: %s' % resp.content)
        return 1

    raw_crash = resp.json()

    # If there's an error in the raw crash, then something is wrong--probably with the API
    # token. So print that out and exit.
    if 'error' in raw_crash:
        print('Error fetching raw crash: %s' % raw_crash['error'], file=sys.stderr)
        return 1

    resp = fetch('/ProcessedCrash/', crash_id, api_token)
    if resp.status_code == 404:
        printerr('%s: does not have processed crash.' % crash_id)
        return 1
    if resp.status_code == 429:
        printerr('API rate limit reached. %s' % resp.content)
        # FIXME(willkg): Maybe there's something better we could do here. Like maybe wait a
        # few minutes.
        return 1
    if resp.status_code == 500:
        printerr('HTTP 500: %s' % resp.content)
        return 1

    processed_crash = resp.json()

    # If there's an error in the processed crash, then something is wrong--probably with the
    # API token. So print that out and exit.
    if 'error' in processed_crash:
        printerr('Error fetching processed crash: %s' % processed_crash['error'])
        return 1

    crash_data = convert_to_crash_data(raw_crash, processed_crash)

    print(json.dumps(crash_data, indent=2))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pkg_resources import parse_version


# yyyymmdd
__releasedate__ = '20180803'

# x.y.z or x.y.z.dev0 -- semver
__version__ = '0.1.3'
VERSION = parse_version(__version__)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys

from .rules import (
    SignatureGenerationRule,
    StackwalkerErrorSignatureRule,
    OOMSignature,
    AbortSignature,
    SignatureShutdownTimeout,
    SignatureRunWatchDog,
    SignatureIPCChannelError,
    SignatureIPCMessageName,
    SigFixWhitespace,
    SigTruncate,
    SignatureJitCategory,
    SignatureParentIDNotEqualsChildID,
)


DEFAULT_PIPELINE = [
    SignatureGenerationRule(),
    StackwalkerErrorSignatureRule(),
    OOMSignature(),
    AbortSignature(),
    SignatureShutdownTimeout(),
    SignatureRunWatchDog(),
    SignatureIPCChannelError(),
    SignatureIPCMessageName(),
    SignatureParentIDNotEqualsChildID(),
    SignatureJitCategory(),

    # NOTE(willkg): These should always come last and in this order
    SigFixWhitespace(),
    SigTruncate(),
]


class SignatureGenerator:
    def __init__(self, pipeline=None, error_handler=None, debug=False):
        """
        :arg pipeline: list of rules to use for signature generation
        :arg error_handler: error handling function with signature
            ``fun(signature_data, exc_info, extra)``
        :arg debug: whether or not to be in debug mode which shows verbose
            output about what happend

        """
        self.pipeline = pipeline or list(DEFAULT_PIPELINE)
        self.error_handler = error_handler
        self.debug = debug

    def generate(self, signature_data):
        """Takes data and returns a signature

        :arg dict signature_data: data to use to generate a signature

        :returns: dict containing ``signature`` and ``notes`` keys representing the
            signature and processor notes

        """
        # NOTE(willkg): Rules mutate the result structure in-place
        result = {
            'signature': '',
            'notes': []
        }

        for rule in self.pipeline:
            rule_name = rule.__class__.__name__

            try:
                if rule.predicate(signature_data, result):
                    old_sig = result['signature']
                    rule.action(signature_data, result)

                    if self.debug:
                        result['notes'].append(
                            '%s: %s -> %s' % (rule_name, old_sig, result['signature'])
                        )

            except Exception as exc:
                if self.error_handler:
                    self.error_handler(
                        signature_data,
                        exc_info=sys.exc_info(),
                        extra={'rule': rule_name}
                    )

                result['notes'].append('Rule %s failed: %s' % (rule_name, exc))

        return result



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import codecs
import os
import re

from setuptools import setup, find_packages


def get_long_desc():
    return codecs.open('README.rst', encoding='utf-8').read()


def get_version():
    fn = os.path.join('siggen', '__init__.py')
    vsre = r"""^__version__ = ['"]([^'"]*)['"]"""
    version_file = codecs.open(fn, mode='r', encoding='utf-8').read()
    return re.search(vsre, version_file, re.M).group(1)


install_requires = [
    'glom',
    'requests',
    'six',
    'ujson',
]


setup(
    name='siggen',
    version=get_version(),
    description='Experimental extraction of Socorro signature generation',
    long_description=get_long_desc(),
    maintainer='Will Kahn-Greene',
    maintainer_email='willkg@mozilla.com',
    url='https://github.com/willkg/socorro-siggen',
    license='Mozilla Public License v2',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    entry_points="""
        [console_scripts]
        signify=siggen.cmd_signify:main
        fetch-data=siggen.cmd_fetch_data:main
        signature=siggen.cmd_signature:main
    """,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
    ]
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function

import argparse
import csv
import os
import sys

import requests

from .generator import SignatureGenerator
from .utils import convert_to_crash_data


DESCRIPTION = """
Given one or more crash ids via command line or stdin (one per line), pulls down information from
Socorro, generates signatures, and prints signature information.
"""

EPILOG = """
Note: In order for the SignatureJitCategory rule to work, you need a valid API token from
Socorro that has "View Personally Identifiable Information" permission.
"""

# FIXME(willkg): This hits production. We might want it configurable.
API_URL = 'https://crash-stats.mozilla.com/api'


class OutputBase:
    """Base class for outputter classes

    Outputter classes are context managers. If they require start/top or begin/end semantics, they
    should implement ``__enter__`` and ``__exit__``.

    Otherwise they can just implement ``data`` and should be fine.

    """
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def warning(self, line):
        """Prints out a warning line to stderr

        :arg str line: the line to print to stderr

        """
        print('WARNING: %s' % line, file=sys.stderr)

    def data(self, crash_id, old_sig, new_sig, notes):
        """Outputs a data point

        :arg str crash_id: the crash id for the signature generated

        :arg str old_sig: the old signature retrieved in the processed crash

        :arg str new_sig: the new generated signature

        :arg list notes: any processor notes

        """
        pass


class TextOutput(OutputBase):
    def data(self, crash_id, old_sig, new_sig, notes):
        print('Crash id: %s' % crash_id)
        print('Original: %s' % old_sig)
        print('New:      %s' % new_sig)
        print('Same?:    %s' % (old_sig == new_sig))

        if notes:
            print('Notes:    (%d)' % len(notes))
            for note in notes:
                print('          %s' % note)


class CSVOutput(OutputBase):
    def __enter__(self):
        self.out = csv.writer(sys.stdout, quoting=csv.QUOTE_ALL)
        self.out.writerow(['crashid', 'old', 'new', 'same?', 'notes'])
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.out = None

    def data(self, crash_id, old_sig, new_sig, notes):
        self.out.writerow([crash_id, old_sig, new_sig, str(old_sig == new_sig), notes])


def fetch(endpoint, crash_id, api_token=None):
    kwargs = {
        'params': {
            'crash_id': crash_id
        }
    }
    if api_token:
        kwargs['headers'] = {
            'Auth-Token': api_token
        }

    return requests.get(API_URL + endpoint, **kwargs)


def main(argv=None):
    """Takes crash data via args and generates a Socorro signature

    """
    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG)
    parser.add_argument(
        '-v', '--verbose', help='increase output verbosity', action='store_true'
    )
    parser.add_argument(
        '--format', help='specify output format: csv, text (default)'
    )
    parser.add_argument(
        '--different-only', dest='different', action='store_true',
        help='limit output to just the signatures that changed',
    )
    parser.add_argument(
        'crashids', metavar='crashid', nargs='*', help='crash id to generate signatures for'
    )

    if argv is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(argv)

    if args.format == 'csv':
        outputter = CSVOutput
    else:
        outputter = TextOutput

    api_token = os.environ.get('SOCORRO_API_TOKEN', '')

    generator = SignatureGenerator(debug=args.verbose)
    if args.crashids:
        crashids_iterable = args.crashids
    elif not sys.stdin.isatty():
        # If a script is piping to this script, then isatty() returns False. If
        # there is no script piping to this script, then isatty() returns True
        # and if we do list(sys.stdin), it'll block waiting for input.
        crashids_iterable = list(sys.stdin)
    else:
        crashids_iterable = []

    if not crashids_iterable:
        parser.print_help()
        return 0

    with outputter() as out:
        for crash_id in crashids_iterable:
            crash_id = crash_id.strip()

            resp = fetch('/RawCrash/', crash_id, api_token)
            if resp.status_code == 404:
                out.warning('%s: does not exist.' % crash_id)
                continue
            if resp.status_code == 429:
                out.warning('API rate limit reached. %s' % resp.content)
                # FIXME(willkg): Maybe there's something better we could do here. Like maybe wait a
                # few minutes.
                return 1
            if resp.status_code == 500:
                out.warning('HTTP 500: %s' % resp.content)
                continue

            raw_crash = resp.json()

            # If there's an error in the raw crash, then something is wrong--probably with the API
            # token. So print that out and exit.
            if 'error' in raw_crash:
                out.warning('Error fetching raw crash: %s' % raw_crash['error'])
                return 1

            resp = fetch('/ProcessedCrash/', crash_id, api_token)
            if resp.status_code == 404:
                out.warning('%s: does not have processed crash.' % crash_id)
                continue
            if resp.status_code == 429:
                out.warning('API rate limit reached. %s' % resp.content)
                # FIXME(willkg): Maybe there's something better we could do here. Like maybe wait a
                # few minutes.
                return 1
            if resp.status_code == 500:
                out.warning('HTTP 500: %s' % resp.content)
                continue

            processed_crash = resp.json()

            # If there's an error in the processed crash, then something is wrong--probably with the
            # API token. So print that out and exit.
            if 'error' in processed_crash:
                out.warning('Error fetching processed crash: %s' % processed_crash['error'])
                return 1

            old_signature = processed_crash['signature']
            crash_data = convert_to_crash_data(raw_crash, processed_crash)

            ret = generator.generate(crash_data)

            if not args.different or old_signature != ret['signature']:
                out.data(crash_id, old_signature, ret['signature'], ret['notes'])
